/* 10/07/2003
  VagueDenoiser plugin for Avisynth -- a wavelet denoiser plugin
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  The author can be contacted at:
  lefungus@altern.org

  Thanks to Geoff Davis for his wavelet Transform Coder Construction Kit
  Look at http://www.geoffdavis.net/ for the full package.
  Thanks to everyone at Doom9.org for their help.

  V0.12: 1st release
  -Precision problems corrected, 0 for threshold is now lossless

  V0.13:
  -Hard thresholding method enabled
  -Cosmetic changes
  -Avisynth parameters changed ( VagueDenoiser(threshold=0.8,wavelet=[0,1],method=[0,1]
   
  V0.2:
  -Implemented many new wavelet filters, thanks to the wavelet transform coder construction kit
   look at http://www.geoffdavis.net/ for more informations
  -Implemented nsteps parameter, that allow you to use n steps in the selected wavelet tranform
   Some nsteps values could produce unvalid results, reducing this value generally solve the problem
  -Little optimizations from Shodan and Bidoche

  V0.22:
  -Implemented YUY2 colorspace
  -Optionnal chroma filtering, (chroma=true/false)
  -Cosmetic changes, code cleaned
  -Html documentation
*/

#include <windows.h>
#include "avisynth.h"
#include <malloc.h>
#include "VagueDenoiser.h"
#include "math.h"

#define sgn(x) ((x<0)?-1:((x>0)?1:0))

VagueDenoiser::VagueDenoiser(PClip _child1, float _threshold, int _wavetype, int _method, int _nsteps, bool _chroma, IScriptEnvironment* env) :
  GenericVideoFilter(_child1),  thresh(_threshold), wavetype(_wavetype), method(_method), nsteps(_nsteps), chroma(_chroma) {
  
  //Wavelet init
  switch(wavetype) {
  case 0: wavelet = new Wavelet(&Haar);
    break;
  case 1: wavelet = new Wavelet(&Daub4);
    break;
  case 2: wavelet = new Wavelet(&Daub6);
    break;
  case 3: wavelet = new Wavelet(&Daub8);
    break;
  case 4: wavelet = new Wavelet(&Antonini);
    break;
  case 5: wavelet = new Wavelet(&Adelson);
    break;
  case 6: wavelet = new Wavelet(&Brislawn);
    break;
  case 7: wavelet = new Wavelet(&Brislawn2);
    break;
  case 8: wavelet = new Wavelet(&Villa5);
    break;
  case 9: wavelet = new Wavelet(&Villa4);
    break;
  case 10: wavelet = new Wavelet(&Villa3);
    break;
  case 11: wavelet = new Wavelet(&Villa6);
    break;
  case 12: wavelet = new Wavelet(&Villa1);
    break;
  case 13: wavelet = new Wavelet(&Villa);
    break;
  case 14: wavelet = new Wavelet(&Villa2);
    break;
  case 15: wavelet = new Wavelet(&Odegard);
  }
  
  width=vi.width;
  height=vi.height;
  widthU=width/2;
  block= (float*) _aligned_malloc(vi.height*vi.width*sizeof(float), 128);
  size_Y=width*height;
  
  //YV12 init
  if (vi.IsYV12()) {
    heightU=height/2;
  } else //YUY2 init 
    if (vi.IsYUY2()) {
    heightU=height;
    blockU= (float*) _aligned_malloc(heightU*widthU*sizeof(float), 128);
    blockV= (float*) _aligned_malloc(heightU*widthU*sizeof(float), 128);
  }
  else env->ThrowError("VagueDenoiser: YV12 or YUY2 only");
 
  size_UV=widthU*heightU;
}

VagueDenoiser::~VagueDenoiser() {
  _aligned_free(block);
  if(vi.IsYUY2()){
    _aligned_free(blockU);
    _aligned_free(blockV);
  }
  delete wavelet;
} 

unsigned char VagueDenoiser::realToChar (float x) { 
  return (x>255?255:(x<0?0:(unsigned char)(x+0.5))); };

//Apply a filtering on wavelet transformed coefficients
void VagueDenoiser::filter(float* block, int b) {
  int i;
  switch(method) {
  case 0://Hard thresholding
    for(i=0;i<b;i++){
      if (fabs(*(block+i)) <= thresh) {
	*(block+i)=0.0f;
      }
    }
    break;
  case 1://Soft thresholding
    for(i=0;i<b;i++){
      float temp=fabs(*(block+i));
      if (temp <= thresh) {
	*(block+i)=0.0f;
      } else {
	*(block+i)= sgn(*(block+i))*(temp-thresh);
      }
    }
  }
}

//Wavelet Transform
void VagueDenoiser::DWTTransform(float* block, int height, int width, boolean inverse) {
  if(!inverse) {
    wavelet->transform2d(block,block,width,height,nsteps);
  } else {
    wavelet->invert2d(block,block,width,height,nsteps);
  }
}

void VagueDenoiser::filterBlockYV12(float* block,int height, int width, const BYTE* src, unsigned char* dst, int srcPitch, int dstPitch, int length) {
  int i,j;    
  float* blockptr=block;

  //Copy Source to Buffer
  for (i=0; i < height;i++) {
    for (j = 0; j < width; j+=2) {
      *(blockptr+j) =  *(src + j);
      *(blockptr+j+1)= *(src+j+1);
    }
    src += srcPitch;
    blockptr += width;
  }

  //Apply filtering
  DWTTransform(block,height,width,false);
  filter(block,size_Y);
  DWTTransform(block,height,width,true);

  //Copy Buffer to Destination
  blockptr=block;
  for (i=0; i < height;i++) {
    for (j = 0; j < width; j+=2) {
      *(dst +j) = realToChar(*(blockptr+j));
      *(dst+j+1)= realToChar(*(blockptr+j+1));
    }
    dst += dstPitch;
    blockptr += width;
  }
}

void VagueDenoiser::filterBlockYUY2(float* block, float* blockU, float* blockV, int height, int width, const BYTE* src, unsigned char* dst, int srcPitch, int dstPitch, int length) {
  int i,j;    
  float* blockptr=block;
  float* blockptrU=blockU;
  float* blockptrV=blockV;

  //Copy Source to Buffer
  for (i=0; i < height;i++) {
    int k=0;
    for (j = 0; j < 4*widthU; j+=4) {
      *(blockptr+(2*k))  = *(src+j);
      *(blockptr+(2*k)+1)= *(src+j+2);
      *(blockptrU+k) = *(src+j+1);
      *(blockptrV+k) = *(src+j+3);
      k++;
    }
    src += srcPitch;
    blockptr += width;
    blockptrU+= widthU;
    blockptrV+= widthU;
  }

  //Apply filtering
  DWTTransform(block,height,width,false);
  filter(block,size_Y);
  DWTTransform(block,height,width,true);

  //Optionnal chroma filtering
  if(chroma) {

    //Filter U Plane
    DWTTransform(blockU,heightU,widthU,false);
    filter(blockU,size_UV);
    DWTTransform(blockU,heightU,widthU,true);
    
    //Filter V Plane
    DWTTransform(blockV,heightU,widthU,false);
    filter(blockV,size_UV);
    DWTTransform(blockV,heightU,widthU,true);
  }
  
  //Copy Buffer to Destination
  blockptr=block;
  blockptrU=blockU;
  blockptrV=blockV;
  for (i=0; i < height;i++) {
    int k=0;
    for (j = 0; j < widthU; j++) {
      *(dst+k)  = realToChar(*(blockptr+(k/2)));
      *(dst+k+2)= realToChar(*(blockptr+(k/2)+1));
      *(dst+k+1)= realToChar(*(blockptrU+j));
      *(dst+k+3)= realToChar(*(blockptrV+j));
      k+=4;
    }
    dst += dstPitch;
    blockptr += width;
    blockptrU+= widthU;
    blockptrV+= widthU;
  }
}
  
PVideoFrame __stdcall VagueDenoiser::GetFrame(int n, IScriptEnvironment* env) {

  PVideoFrame src = child->GetFrame(n, env);
  PVideoFrame dst = env->NewVideoFrame(vi);

  const BYTE* srcp  = src->GetReadPtr();
  const int src_pitch  = src->GetPitch();
  
  unsigned char* dstp  = dst->GetWritePtr();
  const int dst_pitch  = dst->GetPitch();

  if(vi.IsYV12()) {
    const int src_pitchU = src->GetPitch(PLANAR_U);
    const int dst_pitchU = dst->GetPitch(PLANAR_U);
    
    //Filter Y Plane
    filterBlockYV12(block,height,width,srcp,dstp,src_pitch,dst_pitch,size_Y);
    
    //Optionnal chroma filtering
    if(chroma){
      
      //Filter U Plane
      srcp  = src->GetReadPtr(PLANAR_U);
      dstp  = dst->GetWritePtr(PLANAR_U);
      filterBlockYV12(block,heightU,widthU,srcp,dstp,src_pitchU,dst_pitchU,size_UV);
      
      //Filter V Plane
      srcp = src->GetReadPtr(PLANAR_V);
      dstp = dst->GetWritePtr(PLANAR_V);
      filterBlockYV12(block,heightU,widthU,srcp,dstp,src_pitchU,dst_pitchU,size_UV);
      
    } else {
      int i,j;
      
      //Copy U Plane to Destination
      srcp  = src->GetReadPtr(PLANAR_U);
      dstp  = dst->GetWritePtr(PLANAR_U);
      for (i=0; i < heightU;i++) {
	for (j = 0; j < widthU; j+=2) {
	  *(dstp+j)=*(srcp+j);
	  *(dstp+j+1)=*(srcp+j+1);
	}
	dstp += dst_pitchU;
	srcp += src_pitchU;
      }
      
      //Copy V Plane to Destination
      srcp = src->GetReadPtr(PLANAR_V);
      dstp = dst->GetWritePtr(PLANAR_V);
      for (i=0; i < heightU;i++) {
	for (j = 0; j < widthU; j+=2) {
	  *(dstp+j)=*(srcp+j);
	  *(dstp+j+1)=*(srcp+j+1);
	}
	dstp += dst_pitchU;
	srcp += src_pitchU;
      }
    }
  } else {
      filterBlockYUY2(block,blockU,blockV,height,width,srcp,dstp,src_pitch,dst_pitch,size_Y);
  }
  
  return dst;
} 

AVSValue __cdecl Create_VagueDenoiser(AVSValue args, void* user_data, IScriptEnvironment* env) {
  return new VagueDenoiser(args[0].AsClip(), args[1].AsFloat(0), args[2].AsInt(0), args[3].AsInt(0), args[4].AsInt(0), args[5].AsBool(), env);  
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env) {
  env->AddFunction("VagueDenoiser", "c[threshold]f[filter]i[method]i[nsteps]i[chroma]b", Create_VagueDenoiser, 0);
  return "`VagueDenoiser' VagueDenoiser plugin";
}
